from zerohash import util


class ZerohashObject(dict):
    def __init__(self, id=None, credentials=None, **params):
        super().__init__()

        for k, v in params.items():
            self.__setattr__(k, v)

        self.__setattr__("credentials", credentials)
        if id:
            self["id"] = id

    @classmethod
    def construct_from(cls, values, credentials):
        instance = cls(values.get("id"), credentials=credentials)
        instance.refresh_from(values, credentials)
        return instance

    def refresh_from(self, values, credentials=None, partial=False):
        self.credentials = credentials or getattr(values, "credentials", None)

        # TODO: conditional for if partial=True

        # Any values not set in values will be cleared/removed from the object returned after refresh
        removed = set(self.keys()) - set(values)

        # self._transient_values = self._transient_values - set(values)

        for k, v in values.items():
            super().__setitem__(k, util.convert_to_zerohash_object(v, credentials))



import json


class ZerohashResponse:
    def __init__(self, body, code, headers):

        self.body = body
        self.code = code
        self.headers = headers
        json_body = json.loads(body)
        self.data = json_body.get("message") or json_body

        if "page" in json_body:
            self.page = json_body["page"]

        if "total_pages" in json_body:
            self.total_pages = json_body["total_pages"]



from zerohash.resources.digital_asset_addresses import DigitalAssetAddresses
from zerohash.resources.error_object import ErrorObject
from zerohash.resources.liquidity_execute import LiquidityExecute
from zerohash.resources.liquidity_rfq import LiquidityRfq
from zerohash.resources.participants_customers_new import ParticipantCustomerNew
from zerohash.resources.trades_batch import TradesBatch



from zerohash import util
from zerohash.api_requestor import APIRequestor
from zerohash.resources.abstract.api_resource import APIResource


class LiquidityRfq(APIResource):
    OBJECT_NAME = "liquidity.rfq"

    @classmethod
    def retrieve(cls, **kwargs):
        """
        Override the retrieve method to look like the list method as there is no ID to query by in this case.

        """

        requestor = APIRequestor()

        url = cls.class_url()
        response = requestor.request("GET", url, kwargs)
        zerohash_object = util.convert_to_zerohash_object(response)

        return zerohash_object



import datetime
import hashlib
import hmac
import json
import urllib
import uuid
from base64 import b64decode, b64encode
from urllib.parse import urljoin

import zerohash
from zerohash import error, error_classes, http_client
from zerohash.zerohash_response import ZerohashResponse


class APIRequestor:
    def __init__(
        self,
        credentials=None,
        api_base=None,
        client=None,
    ):
        self.api_base = api_base or zerohash.api_base
        self.credentials = credentials or zerohash.credentials

        if client:
            self._client = client
        elif zerohash.default_http_client:
            self._client = zerohash.default_http_client
        else:
            """If no default http client is set, set one to avoid creating one for every request"""
            zerohash.default_http_client = http_client.new_default_http_client()
            self._client = zerohash.default_http_client

    def handle_error_response(self, rbody, rcode, resp, rheaders):
        try:
            err = self.specific_api_error(rbody, rcode, resp, rheaders)
        except (KeyError, TypeError):
            raise error.APIError(
                f"Invalid response from Zero Hash API: {rbody}. HTTP response code {rcode}",
                rbody,
                rcode,
                resp,
            )
        raise err

    def specific_api_error(self, rbody, rcode, resp, rheaders):

        # api_error_code = resp["code"]
        if 400 <= rcode < 500:
            return zerohash.error.ClientError(
                body=resp,  # resp is serialized, rbody is a bytes string of the body b""
                headers=rheaders,
                status_code=rcode,
            )
        elif rcode > 500:
            return zerohash.error.APIError(headers=rheaders, status_code=rcode)
        else:
            return zerohash.error.UnknownError(body=resp, status_code=rcode)
        # try:
        #     return error_classes.ERROR_CLASSES[api_error_code](
        #         code=resp["code"],
        #         body=resp,
        #         status_code=rcode,
        #         headers=rheaders,
        #     )
        # except KeyError:
        #     return error.UnknownError(body=resp, status_code=rcode)

    def interpret_response(self, rbody, rcode, rheaders):
        try:

            resp = ZerohashResponse(rbody, rcode, rheaders)
        except Exception:
            raise Exception(
                f"Invalid response from API: {rbody} (HTTP response code: {rcode})",
                rbody,
                rcode,
                rheaders,
            )
        if not 200 <= rcode < 300:
            self.handle_error_response(rbody, rcode, resp.data, rheaders)
        return resp

    def _sign(
        self, private_key: str, method: str, route: str, json_body: str, timestamp: str
    ) -> bytes:
        """Given a key and data, create and sign a payload.

        :param api_key: Key to sign the message with
        :param method: HTTP method
        :param route: Relative route. EX. /fills
        :param json_body: JSON as a string. Usually created via json.dumps(dict)
        :param timestamp: Unix Epoch time as a string
        :return: Base64 encoded digest
        """
        msg = bytes(timestamp + method.upper() + route + json_body, encoding="utf-8")
        hm = hmac.new(key=b64decode(private_key), msg=msg, digestmod=hashlib.sha256)
        return b64encode(hm.digest())

    def request_headers(self, credentials, method, route, json_body):
        referrer = "https://api.getlinus.io"
        user_agent = "Zerohash Python Library"
        timestamp = str(int(datetime.datetime.now().timestamp()))
        headers = {
            "X-SCX-API-KEY": credentials.public_key,
            "X-SCX-SIGNED": self._sign(
                private_key=credentials.private_key,
                method=method,
                route=route,
                json_body=json_body,
                timestamp=timestamp,
            ),
            "X-SCX-TIMESTAMP": timestamp,  # Unix Epoch
            "X-SCX-PASSPHRASE": credentials.passphrase,
            "Content-Type": "application/json",
        }

        return headers

    def request(self, method, url, params={}, headers=None):
        rbody, rcode, rheaders = self.request_raw(method, url, params)
        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp

    def request_raw(self, method, url, params={}, supplied_headers=None):
        if self.credentials:
            credentials = self.credentials
        else:
            from zerohash import credentials

            credentials = credentials

        if credentials is None:
            raise error.MalformedAuthorizationError("Missing credentials")

        abs_url = "%s%s" % (self.api_base, url)
        if method.lower() in (
            "post",
            "put",
            "patch",
        ):
            post_data = json.dumps(params, separators=(",", ":"))
        elif method.lower() in (
            "get",
            "delete",
        ):
            post_data = json.dumps({}, separators=(",", ":"))
            if params:
                new_path = "?" + urllib.parse.urlencode(params)
                abs_url = abs_url + new_path
                url = url + new_path
                # abs_url = abs_url + "?" + urllib.parse.urlencode(params)

        headers = self.request_headers(
            credentials, method, route=url, json_body=post_data
        )

        rbody, rcode, rheaders = self._client.request_with_retries(
            method, abs_url, headers, post_data
        )

        return rbody, rcode, rheaders



# REFERENCE CODE FROM THE API DOCS
import hashlib
import hmac
import json
from base64 import b64decode, b64encode
from datetime import datetime
from logging import getLogger
from typing import Any, Dict, Optional
from urllib.parse import urljoin

import requests
from dotenv import find_dotenv, load_dotenv

logger = getLogger(__name__)
import os

# NB: THESE CREDENTIALS SHOULD NOT BE STORED IN PLAINTEXT
# Keys here are kept in plaintext for the purposes of demonstration
# We encourage you to encrypt your keys and decrypt them only when being used
load_dotenv(find_dotenv())
URL_BASE = "api.cert.zerohash.com"
HTTP_BASE = "https://" + URL_BASE
API_PUBLIC_KEY = os.environ["PUBLIC_KEY"]  # "usjHuLksaeBXWSsa8uU7ES"
API_PRIVATE_KEY = os.environ[
    "PRIVATE_KEY"
]  # 2mC4ZvVd4goRkuJm+rjr9byUiaUW1b6tVN4xy9QXNSE=
PASSPHRASE = os.environ["PASSPHRASE"]  # testingisgreat


def sign(
    api_key: str, method: str, route: str, json_body: str, timestamp: str
) -> bytes:
    """Given a key and data, create and sign a payload.

    :param api_key: Key to sign the message with
    :param method: HTTP method
    :param route: Relative route. EX. /fills
    :param json_body: JSON as a string. Usually created via json.dumps(dict)
    :param timestamp: Unix Epoch time as a string
    :return: Base64 encoded digest
    """
    msg = bytes(timestamp + method + route + json_body, encoding="utf-8")
    hm = hmac.new(key=b64decode(api_key), msg=msg, digestmod=hashlib.sha256)
    return b64encode(hm.digest())


def headers() -> Dict[str, Any]:
    """Create a header template for use in HTTP requests."""
    return {
        "X-SCX-API-KEY": API_PUBLIC_KEY,
        "X-SCX-SIGNED": "",  # Put here to make sure we alway send something
        # The datetime.timestamp function is available only in Python 3.3+
        "X-SCX-TIMESTAMP": str(int(datetime.now().timestamp())),  # Unix Epoch
        "X-SCX-PASSPHRASE": PASSPHRASE,
    }


def make_seed_request(
    method: str, url: str, body: Optional[Dict[str, str]] = None
) -> requests.Response:
    """Create and send an HTTP request with a signature to the Zero Hash API.

    :param method: HTTP method
    :param url: Relative route. EX. /fills
    :param body: Dictionary for serializing into the JSON body of the request. For GET requests,
                 this can be omitted or set to an empty dict. Nothing will be sent, but it is
                 required for the signature.
    :return: requests.Response object
    """
    if body is None:
        body = {}
    h = headers()
    json_body = json.dumps(body, separators=(",", ":"))
    h["X-SCX-SIGNED"] = sign(
        API_PRIVATE_KEY, method, url, json_body, h["X-SCX-TIMESTAMP"]
    )
    args = {"method": method, "url": urljoin(HTTP_BASE, url)}
    logger.info("Making {} request to {}".format(method, urljoin(URL_BASE, url)))
    if body:
        args["data"] = json_body
        h["Content-Type"] = "application/json"
        logger.debug(json_body)
    args["headers"] = h
    print(args)
    # Since we don't know if it's a GET or POST, use the generic request function and create an
    # args dict so that we can conditionally pass data/JSON
    return requests.request(**args)


def make_seed_reqauest_hardcoded():
    body = {}
    h = {
        "X-SCX-API-KEY": "dhMsj1QcGP3TsepKPRBRcW",
        "X-SCX-SIGNED": "",  # Put here to make sure we alway send something
        # The datetime.timestamp function is available only in Python 3.3+
        "X-SCX-TIMESTAMP": "1633116917",
        "X-SCX-PASSPHRASE": "thisiscool",
    }



import textwrap

import zerohash


class ZerohashError(Exception):
    def __init__(
        self,
        messages=None,
        errors=None,
        body=None,
        status_code=None,
        headers=None,
    ):
        # self._messages = messages
        # self.errors = errors
        self._body = body
        self.status_code = status_code
        self.headers = headers or {}
        self.request_id = self.headers.get("X-Request-Id", None)

        self.messages = self.construct_messages()
        self.errors = self.construct_errors()

    def construct_messages(self):

        # if self._body is None or "messages" not in self._body:
        #     return self._messages

        # messages = getattr(self._body, "messages", self._messages)
        messages = getattr(self._body, "messages", [])
        if isinstance(messages, str):
            return [messages]
        return messages

    def construct_errors(self):

        errors = []

        # Append an error object for each  message
        for msg in self.messages:
            errors.append(
                zerohash.resources.error_object.ErrorObject.construct_from(
                    dict(message=msg), zerohash.credentials
                )
            )

        # Now append an error object for each item in "errors" or the string in "error" in the body
        for msg in getattr(self._body, "errors", []):
            errors.append(
                zerohash.resources.error_object.ErrorObject.construct_from(
                    dict(message=msg), zerohash.credentials
                )
            )
        if hasattr(self._body, "error"):
            errors.append(
                zerohash.resources.error_object.ErrorObject.construct_from(
                    dict(message=self._body["error"]), zerohash.credentials
                )
            )
        return errors

    def __str__(self):
        msg = self.messages
        if self.request_id:
            return f"Request {self.request_id}: {msg}"

        else:
            return msg


class APIError(ZerohashError):
    """Used for 5XX errors received from the Zero Hash API"""

    pass


class ClientError(ZerohashError):
    """Used for 4XX errors received from the Zero Hash API"""

    pass


class UnknownError(ZerohashError):
    pass


class AuthenticationError(ZerohashError):
    pass


class MalformedAuthorizationError(ZerohashError):
    pass


class APIConnectionError(ZerohashError):
    def __init__(
        self,
        message_body,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        should_retry=False,
    ):
        message_body = textwrap.fill(message_body)
        super().__init__(message_body, http_body, http_status, headers, code)
        self.should_retry = should_retry



from zerohash.zerohash_object import ZerohashObject
from zerohash.zerohash_response import ZerohashResponse


def convert_to_zerohash_object(obj, credentials=None):
    """
    For a given object, obj, convert it into the appropriate zerohash object (if applicable)

    If a ZerohashResponse is passed, return the ZerohashObject associated with it

    Otherwise derive a ZerohashObject from the object passed
    """

    if isinstance(obj, ZerohashResponse):
        zerohash_response = obj
        obj = zerohash_response.data
    if isinstance(obj, list):
        return [convert_to_zerohash_object(r, credentials) for r in obj]
    # elif get_object_classes().get(klass_name) is not None:
    #     return get_object_classes().get(klass_name).construct_from(response, credentials)
    elif isinstance(obj, dict) and not isinstance(obj, ZerohashObject):

        return ZerohashObject.construct_from(obj, credentials)
    elif isinstance(obj, dict):
        return ZerohashObject.construct_from(obj, credentials)
    else:
        return obj



__version__ = "0.0.2"  # Do not update directly, use bumpversion

# private_key = None
# public_key = None
# passphrase = None

credentials = None

default_http_client = None
proxy_url = None

api_base = (
    "https://api.cert.zerohash.com"  # set to https://api.zerohash.com in production
)

max_network_retries = 0

from zerohash.resources import *


class Credential:
    def __init__(self, private_key, public_key, passphrase):
        self.private_key = private_key
        self.public_key = public_key
        self.passphrase = passphrase


def set_credentials(private_key, public_key, passphrase):
    return Credential(private_key, public_key, passphrase)



from zerohash import resources

OBJECT_CLASSES = {
    resources.DigitalAssetAddresses.OBJECT_NAME: resources.DigitalAssetAddresses,
    resources.LiquidityRfq.OBJECT_NAME: resources.LiquidityRfq,
    resources.ParticipantCustomerNew.OBJECT_NAME: resources.ParticipantCustomerNew,
    resources.LiquidityExecute.OBJECT_NAME: resources.LiquidityExecute,
    resources.TradesBatch.OBJECT_NAME: resources.TradesBatch,
}



from zerohash import api_requestor, util
from zerohash.resources.abstract.api_resource import APIResource


class CreateableAPIResource(APIResource):
    @classmethod
    def create(cls, credentials=None, idempotency_key=None, **params):

        requestor = api_requestor.APIRequestor(credentials)
        url = cls.class_url()
        response = requestor.request("POST", url, params)
        return util.convert_to_zerohash_object(
            response,
            credentials,  # klass_name=cls.OBJECT_NAME
        )



from zerohash import util
from zerohash.api_requestor import APIRequestor
from zerohash.resources.abstract.api_resource import APIResource


class ListableAPIResource(APIResource):
    @classmethod
    def list(cls, **params):
        requestor = APIRequestor()
        url = cls.class_url()
        response = requestor.request("GET", url, params)
        zerohash_object = util.convert_to_zerohash_object(response)
        return zerohash_object



from zerohash.zerohash_object import ZerohashObject


class APIResource:
    @classmethod
    def retrieve(cls, id, **kwargs):
        instance = cls(id, *kwargs)
        instance.refresh()
        return instance

    @classmethod
    def class_url(cls):
        if cls == APIResource:
            raise NotImplementedError(
                "APIResource is an abstract class. Perform actions on one of its subclasses"
            )

        base = cls.OBJECT_NAME.replace(".", "/")
        return "/%s" % (base,)

    def instance_url(self):
        id = self.get("id")

        base = self.class_url()
        extn = id
        return "%s/%s" % (base, extn)

    def refresh(self):
        self.refresh_from(self.request("get", self.instance_url()))
        return self



from zerohash.resources.abstract import CreateableAPIResource, ListableAPIResource


class DigitalAssetAddresses(ListableAPIResource, CreateableAPIResource):
    OBJECT_NAME = "deposits.digital_asset_addresses"



from zerohash import api_requestor, util
from zerohash.resources.abstract import CreateableAPIResource


class TradesBatch(CreateableAPIResource):
    OBJECT_NAME = "trades.batch"

    @classmethod
    def create(cls, credentials=None, idempotency_key=None, trades=None):
        if trades is None:
            trades = []
        requestor = api_requestor.APIRequestor(credentials)
        url = cls.class_url()
        response = requestor.request("POST", url, params=trades)
        return util.convert_to_zerohash_object(
            response,
            credentials,  # klass_name=cls.OBJECT_NAME
        )
